# -*- coding: utf-8 -*-
# pylint: disable-msg=w0231
'''base stuf'''

from __future__ import absolute_import

from itertools import chain
from collections import Mapping, Sequence, MutableMapping
from operator import getitem, delitem, setitem, methodcaller

from .utils import clsname, lazy, recursive_repr


class corestuf(object):

    '''stuf core stuff'''

    _map = dict
    _reserved = ['allowed', '_wrapped', '_map']

    def __getattr__(self, key):
        try:
            return getitem(self, key)
        except KeyError:
            if key == 'iteritems':
                return self.items
            elif key == 'iterkeys':
                return self.keys
            elif key == 'itervalues':
                return self.values
            return object.__getattribute__(self, key)

    @recursive_repr
    def __repr__(self):
        if not self:
            return '%s()' % clsname(self)
        return '%s(%r)' % (clsname(self), methodcaller('items')(self))

    @lazy
    def _classkeys(self):
        # protected keywords
        try:
            return frozenset(chain(
                vars(self).iterkeys(),
                vars(self.__class__).iterkeys(),
                self._reserved
            ))
        except AttributeError:
            return frozenset(chain(
                vars(self).keys(), vars(self.__class__).keys(), self._reserved
            ))

    @classmethod
    def _build(cls, iterable):
        kind = cls._map
        # add class to handle potential nested objects of the same class
        kw = kind()
        if isinstance(iterable, Mapping):
            try:
                iitems = iterable.iteritems
            except AttributeError:
                iitems = iterable.items
            kw.update(kind(i for i in iitems()))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            for arg in iterable:
                try:
                    kw.update(arg)
                except (ValueError, TypeError):
                    pass
        return kw

    @classmethod
    def _mapping(cls, iterable):
        return cls._map(iterable)

    @classmethod
    def _new(cls, iterable):
        return cls(cls._build(iterable))

    @classmethod
    def _populate(cls, past, future):
        new = cls._new
        try:
            pitems = past.iteritems
            bstring = basestring  # @UndefinedVariable
        except AttributeError:
            pitems = past.items
            bstring = str
        for key, value in pitems():
            if all([
                isinstance(value, (Sequence, Mapping)),
                not isinstance(value, bstring),
            ]):
                # see if stuf can be converted to nested stuf
                trial = new(value)
                if len(trial) > 0:
                    setitem(future, key, trial)
                else:
                    setitem(future, key, value)
            else:
                setitem(future, key, value)
        return cls._postpopulate(future)

    @classmethod
    def _postpopulate(cls, future):
        return future

    def _prepopulate(self, *args, **kw):
        kw.update(self._build(args))
        return kw

    def copy(self):
        return self._build(self._map(self))


class writestuf(corestuf):

    '''stuf basestuf'''

    def __setattr__(self, key, value):
        # handle normal object attributes
        if key == '_classkeys' or key in self._classkeys:
            self.__dict__[key] = value
        # handle special attributes
        else:
            try:
                setitem(self, key, value)
            except:
                raise AttributeError(key)

    def __delattr__(self, key):
        # allow deletion of key-value pairs only
        if not key == '_classkeys' or key in self._classkeys:
            try:
                delitem(self, key)
            except KeyError:
                raise AttributeError(key)

    def __iter__(self):
        cls = self.__class__
        try:
            iitems = self.iteritems
        except AttributeError:
            iitems = self.items
        for key, value in iitems():
            # nested stuf of some sort
            if isinstance(value, cls):
                yield (key, dict(i for i in value))
            # normal key, value pair
            else:
                yield (key, value)

    def __getstate__(self):
        return self._mapping(self)

    def __setstate__(self, state):
        return self._build(state)

    def update(self, *args, **kw):
        self._populate(self._prepopulate(*args, **kw), self)


class directstuf(writestuf):

    '''stuf basestuf'''

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self.update(*args, **kw)


class wrapstuf(corestuf):

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        super(wrapstuf, self).__init__()
        self._wrapped = self._map()
        self._wrapped = self._populate(
            self._prepopulate(*args, **kw), self._wrapped,
        )

    @classmethod
    def _postpopulate(cls, future):
        return cls._mapping(future)


class writewrapstuf(wrapstuf, writestuf, MutableMapping):

    '''wraps mappings for stuf compliance'''

    def __getitem__(self, key):
        return getitem(self._wrapped, key)

    def __setitem__(self, key, value):
        setitem(self._wrapped, key, value)

    def __delitem__(self, key):
        delitem(self._wrapped, key)

    def __iter__(self):
        return iter(self._wrapped)

    def __len__(self):
        return len(self._wrapped)

    def clear(self):
        self._wrapped.clear()


__all__ = ['corestuf', 'directstuf', 'wrapstuf', 'writestuf', 'writewrapstuf']



# -*- coding: utf-8 -*-
## pylint: disable-msg=w0702,f0401
'''stuf utilities'''

from __future__ import absolute_import

try:
    from collections import OrderedDict
except  ImportError:
    from ordereddict import OrderedDict
try:
    from thread import get_ident
except ImportError:
    try:
        from dummy_thread import get_ident
    except ImportError:
        from _thread import get_ident
from functools import wraps, update_wrapper
from operator import itemgetter, attrgetter, getitem, delitem


def attr_or_item(this, key):
    '''
    get attribute or item

    @param this: object
    @param key: key to lookup
    '''
    try:
        return getitem(this, key)
    except KeyError:
        return getter(this, key)


def clsname(this):
    '''
    get class name

    @param this: object
    '''
    return getattr(this.__class__, '__name__')


def deepget(this, key, default=None):
    '''
    get an attribute with deep attribute path

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return attrgetter(key)(this)
    except AttributeError:
        return default


def deleter(this, key):
    '''
    delete an attribute

    @param this: object
    @param key: key to lookup
    '''
    try:
        object.__delattr__(this, key)
    except (TypeError, AttributeError):
        delattr(this, key)


def getcls(this):
    '''
    get class of instance

    @param this: an instance
    '''
    return getter(this, '__class__')


def getter(this, key):
    '''
    get an attribute

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return object.__getattribute__(this, key)
    except (AttributeError, TypeError):
        return getattr(this, key)


def get_or_default(this, key, default=None):
    '''
    get an attribute

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return getter(this, key)
    except AttributeError:
        return default


def instance_or_class(key, this, owner):
    '''
    get attribute of an instance or class

    @param key: name of attribute to look for
    @param this: instance to check for attribute
    @param owner: class to check for attribute
    '''
    try:
        return getter(this, key)
    except AttributeError:
        return getter(owner, key)


def inverse_lookup(value, this, default=None):
    '''
    get attribute of an instance by value

    @param value: value to lookup as a key
    @param this: instance to check for attribute
    @param default: default key (default: None)
    '''
    try:
        return itemgetter(value)(
            dict((v, k) for k, v in vars(this).iteritems())
        )
    except (TypeError, KeyError):
        return default


def lru_wrapped(this, maxsize=100):
    '''
    least-recently-used cache decorator from Raymond Hettinger

    arguments to the cached function must be hashable.

    @param maxsize: maximum number of results in LRU cache (default: 100)
    '''
    # order: least recent to most recent
    cache = OrderedDict()

    @wraps(this)
    def wrapper(*args, **kw):
        key = args
        if kw:
            key += tuple(sorted(kw.items()))
        try:
            result = cache.pop(key)
        except KeyError:
            result = this(*args, **kw)
            # purge least recently used cache entry
            if len(cache) >= maxsize:
                cache.popitem(0)
        # record recent use of this key
        cache[key] = result
        return result
    return wrapper


def selfname(this):
    '''
    get object name

    @param this: object
    '''
    return getter(this, '__name__')


def recursive_repr(this):
    '''
    Decorator to make a repr function return "..." for a recursive call

    @param this: object
    '''
    repr_running = set()

    def wrapper(self):
        key = id(self), get_ident()
        if key in repr_running:
            return '...'
        repr_running.add(key)
        try:
            result = this(self)
        finally:
            repr_running.discard(key)
        return result
    # Can't use functools.wraps() here because of bootstrap issues
    wrapper.__module__ = getattr(this, '__module__')
    wrapper.__doc__ = getattr(this, '__doc__')
    wrapper.__name__ = selfname(this)
    return wrapper


def setter(this, key, value):
    '''
    get an attribute

    @param this: object
    @param key: key to set
    @param value: value to set
    '''
    # it's an instance
    try:
        this.__dict__[key] = value
        return value
    # it's a class
    except TypeError:
        setattr(this, key, value)
        return value


class lazybase(object):

    '''base for lazy descriptors'''

    def __init__(self, method):
        self.method = method
        self.name = selfname(method)
        update_wrapper(self, method)


class lazy(lazybase):

    '''lazily assign attributes on an instance upon first use.'''

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return setter(instance, self.name, self.method(instance))


class lazy_class(lazybase):

    '''Lazily assign attributes on an class upon first use.'''

    def __get__(self, instance, owner):
        return setter(owner, self.name, self.method(owner))


class lazy_set(lazybase):

    '''lazy assign attributes with a custom setter'''

    def __init__(self, method, fget=None):
        super(lazy_set, self).__init__(method)
        self.fget = fget
        update_wrapper(self, method)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return setter(instance, self.name, self.method(instance))

    def __set__(self, this, value):
        self.fget(this, value)

    def __delete__(self, this):
        delitem(this.__dict__, self.name)

    def setter(self, func):
        self.fget = func
        return self


class both(lazybase):

    '''
    decorator which allows definition of a Python descriptor with both
    instance-level and class-level behavior
    '''

    def __init__(self, method, expr=None):
        super(both, self).__init__(method)
        self.expr = expr or method
        update_wrapper(self, method)

    def __get__(self, instance, owner):
        if instance is None:
            return self.expr(owner)
        return setter(instance, self.name, self.method(instance))

    def expression(self, expr):
        '''
        a modifying decorator that defines a general method
        '''
        self.expr = expr
        return self


class either(both):

    '''
    decorator which allows caching results of a Python descriptor with both
    instance-level and class-level behavior
    '''

    def __init__(self, method, expr=None):
        super(either, self).__init__(method)
        self.expr = expr or method
        update_wrapper(self, method)

    def __get__(self, instance, owner):
        if instance is None:
            return setter(owner, self.name, self.expr(owner))
        return setter(instance, self.name, self.method(instance))

    def expression(self, expr):
        '''
        a modifying decorator that defines a general method
        '''
        self.expr = expr
        return self


__all__ = [
    'attr_or_item', 'both', 'clsname', 'deepget', 'deleter', 'either',
    'get_or_default', 'getcls', 'getter', 'instance_or_class',
    'inverse_lookup', 'lazy', 'lazy_class', 'lazy_set', 'lru_wrapped',
    'recursive_repr', 'selfname', 'setter',
]



# -*- coding: utf-8 -*-
'''stuf has attributes'''

from __future__ import absolute_import

from .core import defaultstuf, fixedstuf, frozenstuf, orderedstuf, stuf

idefaultstuf = defaultstuf
ifixedstuf = fixedstuf
ifrozenstuf = frozenstuf
iorderedstuf = orderedstuf
istuf = stuf

__all__ = ['defaultstuf', 'fixedstuf', 'frozenstuf', 'orderedstuf', 'stuf']



# -*- coding: utf-8 -*-
# pylint: disable-msg=w0221,w0212,w0201
'''core stuf'''

from __future__ import absolute_import

from operator import setitem, getitem
from collections import Mapping, Sequence, defaultdict, namedtuple

from .utils import OrderedDict, getter
from .base import directstuf, wrapstuf, writewrapstuf


class defaultstuf(directstuf, defaultdict):

    '''
    dictionary with attribute-style access and a factory function to provide a
    default value for keys with no value
    '''

    _map = defaultdict

    def __getattr__(self, key):
        try:
            if key == 'iteritems':
                return self.items
            elif key == 'iterkeys':
                return self.keys
            elif key == 'itervalues':
                return self.values
            return object.__getattribute__(self, key)
        except AttributeError:
            return getitem(self, key)

    def __init__(self, default, *args, **kw):
        '''
        @param default: function that can provide default values
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        defaultdict.__init__(self, default)
        directstuf.__init__(self, *args, **kw)

    @classmethod
    def _build(cls, default, iterable):
        kind = cls._map
        # add class to handle potential nested objects of the same class
        kw = kind(default)
        if isinstance(iterable, Mapping):
            kw.update(kind(default, iterable))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            for arg in iterable:
                try:
                    kw.update(arg)
                except (ValueError, TypeError):
                    pass
        return kw

    @classmethod
    def _new(cls, default, iterable):
        return cls(default, cls._build(default, iterable))

    def _populate(self, past, future):
        new = self._new
        try:
            pitems = past.iteritems
        except AttributeError:
            pitems = past.items
        for key, value in pitems():
            if isinstance(value, (Mapping, Sequence)):
                # see if stuf can be converted to nested stuf
                trial = new(self.default_factory, value)
                if len(trial) > 0:
                    setitem(future, key, trial)
                else:
                    setitem(future, key, value)
            else:
                setitem(future, key, value)

    def _prepopulate(self, *args, **kw):
        kw.update(self._build(self.default_factory, args))
        return kw

    def copy(self):
        return self._build(self.default_factory, dict(i for i in self))


class fixedstuf(writewrapstuf):

    '''
    dictionary with attribute-style access with mutability restricted to
    initial keys
    '''

    def __setitem__(self, key, value):
        # only access initial keys
        if key in self.allowed:
            super(fixedstuf, self).__setitem__(key, value)
        else:
            raise KeyError('key "{0}" not allowed'.format(key))

    def __reduce__(self):
        return (self.__class__, (self._wrapped.copy(),))

    def _prepopulate(self, *args, **kw):
        iterable = super(fixedstuf, self)._prepopulate(*args, **kw)
        try:
            ikeys = iterable.iterkeys
        except AttributeError:
            ikeys = iterable.keys
        self.allowed = frozenset(ikeys())
        return iterable

    def popitem(self):
        raise AttributeError()

    def pop(self, key, default=None):
        raise AttributeError()


class frozenstuf(wrapstuf, Mapping):

    '''immutable dictionary with attribute-style access'''

    __slots__ = ['_wrapped']

    def __getitem__(self, key):
        try:
            return getattr(self._wrapped, key)
        except AttributeError:
            raise KeyError('key {0} not found'.format(key))

    def __iter__(self):
        try:
            return getter(self, '_wrapped')._asdict().iterkeys()
        except AttributeError:
            return iter(getter(self, '_wrapped')._asdict().keys())

    def __len__(self):
        return len(getter(self, '_wrapped')._asdict())

    def __reduce__(self):
        return (self.__class__, (getter(self, '_wrapped')._asdict().copy(),))

    @classmethod
    def _mapping(self, mapping):
        try:
            mkeys = mapping.iterkeys
        except AttributeError:
            mkeys = mapping.keys
        frozen = namedtuple('frozenstuf', mkeys())
        return frozen(**mapping)


class orderedstuf(writewrapstuf):

    '''dictionary with dot attributes that remembers insertion order'''

    _mapping = OrderedDict

    def __reversed__(self):
        return getter(self, '_wrapped').__reversed__()

    def __reduce__(self):
        return self._wrapped.__reduce__()


class stuf(directstuf, dict):

    '''dictionary with attribute-style access'''


__all__ = ['defaultstuf', 'fixedstuf', 'frozenstuf', 'orderedstuf', 'stuf']
